@@ -1,4 +1,4 @@
-#if defined __PMACCTD_C || defined __NFACCTD_C 
+#if defined __PMACCTD_C || defined __NFACCTD_C || defined __SFACCTD_C 
 #define EXT 
 #else
 #define EXT extern
@@ -11,5 +11,6 @@ EXT u_int32_t NfDataHdrV9Sz, NfTplHdrV9Sz;
 EXT u_int32_t NfDataV1Sz, NfDataV5Sz, NfDataV7Sz;
 EXT u_int32_t IP4HdrSz, IP4TlSz, IP6HdrSz, IP6AddrSz, IP6TlSz; 
 EXT u_int32_t MyTLHdrSz, TCPFlagOff;
+EXT u_int32_t SFSampleSz;
 
 #undef EXT
